import 'package:cruiser/src/errors/route_not_found.dart';
import 'package:cruiser/src/logger/cruiser_logger.dart';
import 'package:cruiser/src/models/arguments_wrapper.dart';
import 'package:cruiser/src/models/cruiser_args.dart';
import 'package:cruiser/src/models/cruiser_options.dart';
import 'package:cruiser/src/models/cruiser_route.dart';
import 'package:cruiser/src/navigator_observers/cruiser_stack_observer.dart';
import 'package:cruiser/src/cruiser_page_route.dart';
import 'package:cruiser/src/enums/transition_type.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';

enum NavigationType { push, pushReplace, pushAndRemoveUntil, popAndPushNamed }

class Cruiser 
{
    final CruiserStackObserver observer = CruiserStackObserver();

    Cruiser({this.options = const CruiserOptions(isLoggingEnabled: true),}) 
    {
        CruiserLogger.init(isLoggerEnabled: options.isLoggingEnabled);

        this._navigatorKey = options.navigatorKey ?? GlobalKey<NavigatorState>();
    }

    final CruiserOptions options;

    Map<String, CruiserRoute> _routeNameMappings = {};

    late GlobalKey<NavigatorState> _navigatorKey;

    GlobalKey<NavigatorState> get navigatorKey => _navigatorKey;


    List<String> getRegisteredRouteNames() => _routeNameMappings.keys.toList();
    
    static T args<T extends CruiserArgs>(BuildContext context) 
    {
        var settings = ModalRoute.of(context)!.settings;

        return (settings.arguments as ArgumentsWrapper).cruiserArgs as T;
    }

    void addRoute(CruiserRoute route) 
    {
        if (_routeNameMappings.containsKey(route.name)) 
        {
            CruiserLogger.instance.warning( "'${route.name}' has already been registered before. Overring it!");
        }

        _routeNameMappings[route.name] = route;
    }

    Future<T?> navigate<T>
    (
        String name, {
        required TransitionType transition,
        CruiserArgs? args,
        NavigationType navigationType = NavigationType.push,
        dynamic result,
        bool Function(Route<dynamic> route)? removeUntilPredicate,
    }) async {
        assert(navigationType != NavigationType.pushAndRemoveUntil || removeUntilPredicate != null);

        _checkAndThrowRouteNotFound(name, args, navigationType);

        final argsWrapper = ArgumentsWrapper(
            cruiserArgs: args,
            transition: transition
        );

        currentTransition = transition;

        switch (navigationType) 
        {
            case NavigationType.push:
            {
                return await this.navigatorKey.currentState!.pushNamed(name, arguments: argsWrapper);
            }
            case NavigationType.pushReplace: 
            {
                return await this.navigatorKey.currentState!.pushReplacementNamed(name, result: result, arguments: argsWrapper);
            }
            case NavigationType.pushAndRemoveUntil:
            {
                return await this.navigatorKey.currentState!.pushNamedAndRemoveUntil<T>(name, removeUntilPredicate!,arguments: argsWrapper);
            }
            case NavigationType.popAndPushNamed:
            {
                return await this.navigatorKey.currentState!.popAndPushNamed(name, result: result, arguments: argsWrapper);
            }
        }
    }

    void _checkAndThrowRouteNotFound(String name, CruiserArgs? args, NavigationType navigationType) 
    {
        if (!_routeNameMappings.containsKey(name)) 
        {
            throw RouteNotFoundError(name: name);
        }
    }

    void pop([dynamic result]) 
    {
        if(observer.routeStack.length <= 1 && !options.allowPopLastRoute)
        {
            return;
        }

        var lastRoute = observer.routeStack.last;

        if(lastRoute is PageRoute)
        {
            currentTransition = _getTransitionFromPageRoute(lastRoute) ?? options.defaultTransition;
        }

        this.navigatorKey.currentState!.pop(result);
    }

    void popUntil(bool Function(Route<dynamic>) predicate)
    {
        // TODO: Set the current transition 
        this.navigatorKey.currentState!.popUntil(predicate);
    }

    Route<dynamic>? generator(RouteSettings settings)
    {
        final route = _routeNameMappings[settings.name];

        if (route == null) return null;

        var argsWrapper = settings.arguments as ArgumentsWrapper? ?? ArgumentsWrapper();

        final CruiserArgs? baseArgs = argsWrapper.cruiserArgs;

        var transition = argsWrapper.transition ?? this.options.defaultTransition;

        RouteSettings routeSettings = RouteSettings
        (
            name: settings.name,
            arguments: argsWrapper.copyWith(cruiserArgs: baseArgs != null ? baseArgs : route.defaultArgs)
        );

        return getPageRoute(route, routeSettings, baseArgs ?? route.defaultArgs, transition);
    } 

    PageRoute getPageRoute(CruiserRoute route, RouteSettings settings, CruiserArgs? args, TransitionType transition)
    {
        switch(transition)
        {
            default:
            {
                return CruiserPageRoute(
                    router: this,
                    settings: settings,
                    builder: (context) => route.builder(context, args),
                    transition: transition
                );
            }

        }
    }

    TransitionType? _getTransitionFromPageRoute(PageRoute pageRoute)
    {
        if(pageRoute is CruiserPageRoute)
        {
            return pageRoute.transition;
        }

        return null;
    }

    TransitionType currentTransition = TransitionType.depth;
}